/*
 *   ALSA driver for VIA VT1732 (Envy24-II)
 *
 *	Copyright (c) 2007 Takashi Iwai <tiwai@suse.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */      

#include <asm/io.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>
#include <linux/moduleparam.h>
#include <linux/mutex.h>
#include <sound/core.h>
#include <sound/info.h>
#include <sound/mpu401.h>
#include <sound/initval.h>
#include <sound/asoundef.h>

#include "vt1732.h"

MODULE_AUTHOR("Takashi Iwai <tiwai@suse.de>");
MODULE_DESCRIPTION("VIA VT1732 (Envy24-II)");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("{{VIA,VT1732}}");;


static int index[SNDRV_CARDS] = SNDRV_DEFAULT_IDX;	/* Index 0-MAX */
static char *id[SNDRV_CARDS] = SNDRV_DEFAULT_STR;	/* ID for this card */
static int enable[SNDRV_CARDS] = SNDRV_DEFAULT_ENABLE_PNP;/* Enable this card */
static char *model[SNDRV_CARDS];

module_param_array(index, int, NULL, 0444);
MODULE_PARM_DESC(index, "Index value for VT1732 soundcard.");
module_param_array(id, charp, NULL, 0444);
MODULE_PARM_DESC(id, "ID string for VT1732 soundcard.");
module_param_array(enable, bool, NULL, 0444);
MODULE_PARM_DESC(enable, "Enable VT1732 soundcard.");
module_param_array(model, charp, NULL, 0444);
MODULE_PARM_DESC(model, "Use the given board model.");

#define PFX "vt1732: "

/*
 */

#define MAX_DMA_BUF_SIZE	(1 << (16 + 2))

/*
 */

static unsigned int rates[] = {
	8000, 9600, 11025, 12000, 16000, 22050, 24000,
	32000, 44100, 48000, 64000, 88200, 96000,
	176400, 192000,
};

static char *rate_texts[] = {
	"8000", "9600", "11025", "12000", "16000", "22050", "24000",
	"32000", "44100", "48000", "64000", "88200", "96000",
	"176400", "192000",
};

#define NUM_RATES_192	ARRAY_SIZE(rates)	/* up to 192k */
#define NUM_RATES_96	(ARRAY_SIZE(rates) - 2)	/* up to 96k */
#define NUM_RATES_48	(ARRAY_SIZE(rates) - 5)	/* up to 48k */

static struct snd_pcm_hw_constraint_list hw_constraints_rates_192 = {
	.count = NUM_RATES_192,
	.list = rates,
};

static struct snd_pcm_hw_constraint_list hw_constraints_rates_96 = {
	.count = NUM_RATES_96,
	.list = rates,
};

static struct snd_pcm_hw_constraint_list hw_constraints_rates_48 = {
	.count = NUM_RATES_48,
	.list = rates,
};

static unsigned int sprates[] = {
	32000, 44100, 48000, 88200, 96000, 176400, 192000,
};

static struct snd_pcm_hw_constraint_list hw_constraints_sprates_192 = {
	.count = ARRAY_SIZE(sprates),
	.list = sprates,
};

static struct snd_pcm_hw_constraint_list hw_constraints_sprates_96 = {
	.count = ARRAY_SIZE(sprates) - 2, /* up to 96000 */
	.list = sprates,
};

static struct snd_pcm_hw_constraint_list hw_constraints_sprates_48 = {
	.count = ARRAY_SIZE(sprates) - 4, /* up to 48000 */
	.list = sprates,
};

static int limit_rate(struct snd_vt1732 *chip,
		      struct snd_pcm_runtime *runtime, int spdif)
{
	if (chip->rate_locked || chip->nopened > 0) {
		if (spdif) {
			switch (chip->rate) {
			case 32000:
			case 44100:
			case 48000:
			case 88200:
			case 96000:
			case 176400:
			case 192000:
				break;
			default:
				return -EINVAL;
			}
		}
		runtime->hw.rates = SNDRV_PCM_RATE_KNOT;
		runtime->hw.rate_min = runtime->hw.rate_max = chip->rate;
	} else if (spdif) {
		switch (chip->rate_max) {
		case 192000:
			snd_pcm_hw_constraint_list(runtime, 0,
						   SNDRV_PCM_HW_PARAM_RATE,
						   &hw_constraints_sprates_192);
			break;
		case 96000:
			snd_pcm_hw_constraint_list(runtime, 0,
						   SNDRV_PCM_HW_PARAM_RATE,
						   &hw_constraints_sprates_96);
			break;
		default:
			snd_pcm_hw_constraint_list(runtime, 0,
						   SNDRV_PCM_HW_PARAM_RATE,
						   &hw_constraints_sprates_48);
			break;
		}
	} else {
		switch (chip->rate_max) {
		case 192000:
			snd_pcm_hw_constraint_list(runtime, 0,
						   SNDRV_PCM_HW_PARAM_RATE,
						   &hw_constraints_rates_192);
			break;
		case 96000:
			snd_pcm_hw_constraint_list(runtime, 0,
						   SNDRV_PCM_HW_PARAM_RATE,
						   &hw_constraints_rates_96);
			break;
		default:
			snd_pcm_hw_constraint_list(runtime, 0,
						   SNDRV_PCM_HW_PARAM_RATE,
						   &hw_constraints_rates_48);
			break;
		}
	}
	return 0;
}

static int snd_vt1732_set_rate(struct snd_vt1732 *chip, unsigned int rate,
			       int force)
{
	unsigned char oval, val;

	if (rate > chip->rate_max)
		return -EINVAL;

	if (!force && chip->rate == rate)
		return 0;

	switch (rate) {
	case 8000: val = 6; break;
	case 9600: val = 3; break;
	case 11025: val = 10; break;
	case 12000: val = 2; break;
	case 16000: val = 5; break;
	case 22050: val = 9; break;
	case 24000: val = 1; break;
	case 32000: val = 4; break;
	case 44100: val = 8; break;
	case 48000: val = 0; break;
	case 64000: val = 15; break;
	case 88200: val = 11; break;
	case 96000: val = 7; break;
	case 176400: val = 12; break;
	case 192000: val = 14; break;
	default:
		snd_BUG();
		return -EINVAL;
	}

	chip->rate = rate;
	spin_lock_irq(&chip->reg_lock);
	oval = vt1732_readb(chip, VT1732_RATE_SELECT);
	val |= oval & ~0x0f;
	if (chip->spdif_in_sync)
		val |= 0x10;
	else
		val &= ~0x10;
	if (val != oval)
		vt1732_writeb(chip, VT1732_RATE_SELECT, val);
	spin_unlock_irq(&chip->reg_lock);

	return 0;
}

/* update spdif control bits; call with reg_lock */
static void update_spdif_bits(struct snd_vt1732 *chip, unsigned int val)
{
	unsigned char cbit, disabled;

	/* disable SPDIF output once */
	cbit = vt1732_readb(chip, VT1732_SPDIF_CONFIG);
	disabled = cbit & ~VT1732_CFG_SPDF_EN_TX;
	if (cbit != disabled)
		vt1732_writeb(chip, VT1732_SPDIF_CONFIG, disabled);
	vt1732_writew(chip, VT1732_SPDF_TX_CTRL, val);
	/* resume SPDIF enable bit */
	if (cbit != disabled)
		vt1732_writeb(chip, VT1732_SPDIF_CONFIG, cbit);
	vt1732_writew(chip, VT1732_SPDF_TX_CTRL, val);
}

/* update SPDIF control bits according to the given rate */
static void update_spdif_rate(struct snd_vt1732 *chip, unsigned int rate)
{
	unsigned int val, nval;
	unsigned long flags;

	spin_lock_irqsave(&chip->reg_lock, flags);
	nval = val = vt1732_readw(chip, VT1732_SPDF_TX_CTRL);
	nval &= ~VT1732_SPDF_TX_RATE;
	switch (rate) {
	case 44100: nval |= VT1732_SPDF_TX_RATE_44; break;
	case 48000: nval |= VT1732_SPDF_TX_RATE_48; break;
	case 32000: nval |= VT1732_SPDF_TX_RATE_32; break;
	case 88200: nval |= VT1732_SPDF_TX_RATE_88; break;
	case 96000: nval |= VT1732_SPDF_TX_RATE_96; break;
	case 192000: nval |= VT1732_SPDF_TX_RATE_192; break;
	case 176400: nval |= VT1732_SPDF_TX_RATE_176; break;
	}
	if (val != nval)
		update_spdif_bits(chip, nval);
	spin_unlock_irqrestore(&chip->reg_lock, flags);
}

static inline int vt1732_check_stream(struct snd_vt1732 *chip,
				      struct snd_pcm_substream *substream,
				      unsigned int idx)
{
	int dir = substream->stream;
	return chip->str[dir][idx].substream != NULL;
}

static void vt1732_reserve_streams(struct snd_vt1732 *chip,
				   struct snd_pcm_substream *substream,
				   unsigned int from, unsigned int to)
{
	int dir = substream->stream;
	unsigned int i;
	for (i = from; i <= to; i++)
		chip->str[dir][i].substream = substream;
	chip->nopened++;
}

static void vt1732_release_streams(struct snd_vt1732 *chip,
				   struct snd_pcm_substream *substream)
{
	int dir = substream->stream;
	unsigned int i;
	for (i = 0; i < VT1732_PCM_STREAMS; i++)
		if (chip->str[dir][i].substream == substream)
			chip->str[dir][i].substream = NULL;
	chip->nopened--;
}

#define vt1732_pcm_running(chip,dir,idx) \
	((chip)->str[dir][idx].running && (chip)->str[dir][idx].substream)


static irqreturn_t vt1732_interrupt(int irq, void *dev_id)
{
	struct snd_vt1732 *chip = dev_id;
	static unsigned int stat_regs[2] = { VT1732_PLAY_INT, VT1732_REC_INT };
	int irq_handled = 0, handled;

	spin_lock(&chip->reg_lock);
	do {
		unsigned int i, j, stat;

		handled = 0;
		stat = vt1732_readw(chip, VT1732_DMA_INT_MASK);
		for (i = 0; i < 2; i++) {
			unsigned int shift = 1 - i;
			if (!(stat & ((VT1732_UART1_RX_INT_STAT |
				       VT1732_UART1_TX_INT_STAT) << shift)))
				continue;
			if (chip->uart[i]) {
				if (stat & (VT1732_UART1_RX_INT_STAT << shift))
					snd_mpu401_uart_interrupt(irq,
								  chip->uart[i]);
				if (stat & (VT1732_UART1_TX_INT_STAT << shift))
					snd_mpu401_uart_interrupt_tx(irq,
								     chip->uart[i]);
				handled = 1;
			} else
				/* invalid irq, mask them */
				stat |= (VT1732_UART1_TX_INT_MASK |
					 VT1732_UART1_RX_INT_MASK) << shift;
		}
		vt1732_writew(chip, VT1732_DMA_INT_MASK, stat);

		for (j = 0; j < 2; j++) {
			stat = vt1732_readl(chip, stat_regs[j]);
			for (i = 0; i < VT1732_PCM_STREAMS; i++) {
				struct snd_vt1732_stream *str;
				str = &chip->str[j][i];
				if (!(stat & str->stat_bit) ||
				    !vt1732_pcm_running(chip, j, i))
					continue;
				spin_unlock(&chip->reg_lock);
				snd_pcm_period_elapsed(str->substream);
				spin_lock(&chip->reg_lock);
				handled = 1;
			}
			vt1732_writel(chip, stat_regs[j], stat);
		}

		irq_handled |= handled;
	} while (handled);
	spin_unlock(&chip->reg_lock);

	return IRQ_RETVAL(irq_handled);
}

static int vt1732_pcm_hw_params(struct snd_pcm_substream *substream,
				struct snd_pcm_hw_params *params)
{
	return snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(params));
}

static int vt1732_pcm_hw_free(struct snd_pcm_substream *substream)
{
	return snd_pcm_lib_free_pages(substream);
}

static inline unsigned char get_format_bit(snd_pcm_format_t fmt)
{
	switch (fmt) {
	case SNDRV_PCM_FORMAT_U8:
		return VT1732_PDMAx_FMT_8BIT;
	case SNDRV_PCM_FORMAT_S16_LE:
		return VT1732_PDMAx_FMT_16BIT;
	default:
		return VT1732_PDMAx_FMT_24BIT;
	}
}

static int vt1732_playback_trigger(struct snd_pcm_substream *substream,
				   int cmd, unsigned int idx)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	unsigned int mask = chip->str[0][idx].stat_bit;
	int err = 0;

	spin_lock(&chip->reg_lock);
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		vt1732_writel(chip, VT1732_PLAY_START,
			      vt1732_readl(chip, VT1732_PLAY_START) | mask);
		vt1732_writel(chip, VT1732_PLAY_INT_MASK,
			      vt1732_readl(chip, VT1732_PLAY_INT_MASK) & ~mask);
		chip->str[0][idx].running = 1;
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		vt1732_writel(chip, VT1732_PLAY_START,
			      vt1732_readl(chip, VT1732_PLAY_START) & ~mask);
		vt1732_writel(chip, VT1732_PLAY_INT_MASK,
			      vt1732_readl(chip, VT1732_PLAY_INT_MASK) | mask);
		chip->str[0][idx].running = 0;
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
		vt1732_writel(chip, VT1732_PLAY_PAUSE,
			      vt1732_readl(chip, VT1732_PLAY_PAUSE) | mask);
		break;
	case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
		vt1732_writel(chip, VT1732_PLAY_PAUSE,
			      vt1732_readl(chip, VT1732_PLAY_PAUSE) & ~mask);
		break;
	default:
		err = -EINVAL;
		break;
	}
	spin_unlock(&chip->reg_lock);
	return err;
}

static int vt1732_capture_trigger(struct snd_pcm_substream *substream,
				  int cmd, unsigned int idx)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	unsigned int mask = chip->str[1][idx].stat_bit;
	int err = 0;

	spin_lock(&chip->reg_lock);
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		vt1732_writel(chip, VT1732_REC_START,
			      vt1732_readl(chip, VT1732_REC_START) | mask);
		vt1732_writel(chip, VT1732_REC_INT_MASK,
			      vt1732_readl(chip, VT1732_REC_INT_MASK) & ~mask);
		chip->str[1][idx].running = 1;
		break;
	case SNDRV_PCM_TRIGGER_STOP:
		vt1732_writel(chip, VT1732_REC_START,
			      vt1732_readl(chip, VT1732_REC_START) & ~mask);
		vt1732_writel(chip, VT1732_REC_INT_MASK,
			      vt1732_readl(chip, VT1732_REC_INT_MASK) | mask);
		chip->str[1][idx].running = 0;
		break;
	default:
		err = -EINVAL;
		break;
	}
	spin_unlock(&chip->reg_lock);
	return err;
}

/*
 * interleaved multi-channel streams
 */

static struct snd_pcm_hardware interleave_pcm_hw = {
	.info =		(SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED |
			 SNDRV_PCM_INFO_BLOCK_TRANSFER |
			 SNDRV_PCM_INFO_MMAP_VALID |
			 SNDRV_PCM_INFO_PAUSE),
	.formats =	(SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S32_LE),
	.rates =	SNDRV_PCM_RATE_KNOT | SNDRV_PCM_RATE_8000_192000,
	.rate_min =	8000,
	.rate_max =	192000,
	.channels_min =		2,
	.channels_max =		16,
	.buffer_bytes_max =	(256*8*1024),
	.period_bytes_min =	2 * 4 * 2,
	.period_bytes_max =	128*8*1024,
	.periods_min =		2,
	.periods_max =		1024,
};

static int vt1732_xdma0_prepare(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct snd_vt1732_stream *str;
	unsigned int base_reg;
	unsigned int bufsize, nbufs, i, addr;
	int err;

	err = snd_vt1732_set_rate(chip, runtime->rate, 0);
	if (err < 0)
		return err;

	str = &chip->str[substream->stream][0];
	base_reg = str->base_reg;

	/* setup PDMA0 buffer */
	bufsize = snd_pcm_lib_buffer_bytes(substream);
	str->buffer_bytes = bufsize;

	nbufs = 0;
	addr = runtime->dma_addr;
	for (i = 0; i < 8; i++) {
		vt1732_writew(chip, base_reg + VT1732_xDMA0_BASE_COUNT_OFS
			      + i * 2, 0);
		vt1732_writel(chip, base_reg + i * 4, 0);
	}
	while (bufsize > 0) {
		unsigned int size;
		size = bufsize;
		if (size > MAX_DMA_BUF_SIZE)
			size = MAX_DMA_BUF_SIZE;
		vt1732_writel(chip, base_reg + nbufs * 4, addr);
		vt1732_writew(chip, base_reg + VT1732_xDMA0_BASE_COUNT_OFS
			      + nbufs * 2, (size >> 2) - 1);
		bufsize -= size;
		addr += size;
		nbufs++;
	}
	str->nbufs = nbufs;
	vt1732_writeb(chip, base_reg + VT1732_xDMA0_FORMAT1_OFS,
		      runtime->channels - 1);
	vt1732_writeb(chip, base_reg + VT1732_xDMA0_FORMAT2_OFS,
		      get_format_bit(runtime->format) | ((nbufs - 1) << 3));
	vt1732_writel(chip, base_reg + VT1732_xDMA0_TERM_COUNT_OFS,
		      (snd_pcm_lib_period_bytes(substream) >> 2) - 1);

	return 0;
}

static snd_pcm_uframes_t
vt1732_xdma0_pointer(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_vt1732_stream *str;
	unsigned int idx, pos, count;

	str = &chip->str[substream->stream][0];
	pos = vt1732_readl(chip, str->base_reg + VT1732_xDMA0_BASE_COUNT_OFS);
	idx = (pos >> 16) & 0x1f;
	if (idx >= str->nbufs) {
		printk(KERN_ERR PFX "invalid PDMA0 index, pos=0x%x\n", pos);
		return 0;
	}
	pos = (pos & 0xffff) << 2;
	if (idx < str->nbufs - 1)
		count = MAX_DMA_BUF_SIZE;
	else
		count = str->buffer_bytes % MAX_DMA_BUF_SIZE;
	if (pos >= count) {
		printk(KERN_ERR PFX "invalid PDMA0 count: 0x%x\n", pos);
		return 0;
	}
	pos = MAX_DMA_BUF_SIZE * idx + (count - 1 - pos);
	return bytes_to_frames(substream->runtime, pos);
}

static int vt1732_pdma0_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_playback_trigger(substream, cmd, 0);
}

static int vt1732_rdma0_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_capture_trigger(substream, cmd, 0);
}


#define VT1732_BUFFER_ALIGN	0x20

static int interleave_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	int dir = substream->stream;
	unsigned int i;

	mutex_lock(&chip->open_mutex);
	runtime->hw = interleave_pcm_hw;
	for (i = 0; i < chip->hw_nstreams[dir]; i++)
		if (vt1732_check_stream(chip, substream, i))
			break;
	if (i < 2) {
		mutex_unlock(&chip->open_mutex);
		return -EBUSY;
	}
	runtime->hw.channels_max = i / 2;
	vt1732_reserve_streams(chip, substream, 0, i - 1);
	limit_rate(chip, runtime, 0);
	snd_pcm_hw_constraint_step(substream->runtime, 0,
				   SNDRV_PCM_HW_PARAM_CHANNELS, 2);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   VT1732_BUFFER_ALIGN);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   VT1732_BUFFER_ALIGN);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static int interleave_pcm_close(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);

	mutex_lock(&chip->open_mutex);
	vt1732_release_streams(chip, substream);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static struct snd_pcm_ops interleave_playback_ops = {
	.open =		interleave_pcm_open,
	.close =	interleave_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	vt1732_xdma0_prepare,
	.trigger =	vt1732_pdma0_trigger,
	.pointer =	vt1732_xdma0_pointer,
};

static struct snd_pcm_ops interleave_capture_ops = {
	.open =		interleave_pcm_open,
	.close =	interleave_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	vt1732_xdma0_prepare,
	.trigger =	vt1732_rdma0_trigger,
	.pointer =	vt1732_xdma0_pointer,
};


/*
 * 2-channel stereo streams
 */

static struct snd_pcm_hardware stereo_pcm_hw = {
	.info =		(SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED |
			 SNDRV_PCM_INFO_BLOCK_TRANSFER |
			 SNDRV_PCM_INFO_MMAP_VALID |
			 SNDRV_PCM_INFO_PAUSE),
	.formats =	(SNDRV_PCM_FMTBIT_U8 |
			 SNDRV_PCM_FMTBIT_S16_LE |
			 SNDRV_PCM_FMTBIT_S32_LE),
	.rates =	SNDRV_PCM_RATE_KNOT | SNDRV_PCM_RATE_8000_192000,
	.rate_min =	8000,
	.rate_max =	192000,
	.channels_min =		2,
	.channels_max =		2,
	.buffer_bytes_max =	(256*2*1024),
	.period_bytes_min =	2 * 4 * 2,
	.period_bytes_max =	256*1024,
	.periods_min =		2,
	.periods_max =		1024,
};

static int stereo_pcm_prepare(struct snd_pcm_substream *substream,
			      unsigned int nstr)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct snd_vt1732_stream *str;
	unsigned int base_reg;
	unsigned int bufsize, nbufs;
	unsigned long addr;
	int err;

	if (nstr == 0)
		return vt1732_xdma0_prepare(substream);

	err = snd_vt1732_set_rate(chip, runtime->rate, 0);
	if (err < 0)
		return err;

	str = &chip->str[substream->stream][nstr];
	base_reg = str->base_reg;

	bufsize = snd_pcm_lib_buffer_bytes(substream);
	str->buffer_bytes = bufsize;

	nbufs = 0;
	addr = runtime->dma_addr;
	while (bufsize > 0) {
		unsigned int reg = base_reg + nbufs * 0x10;
		unsigned int size;
		size = bufsize;
		if (size > MAX_DMA_BUF_SIZE)
			size = MAX_DMA_BUF_SIZE;
		vt1732_writel(chip, reg + VT1732_PDMAx_BASE_ADDR, addr);
		vt1732_writew(chip, reg + VT1732_PDMAx_BASE_COUNT,
			      (size >> 2) - 1);
		bufsize -= size;
		addr += size;
		nbufs++;
	}
	str->nbufs = nbufs;
	vt1732_writeb(chip, base_reg + VT1732_PDMAx_FORMAT,
		      (nbufs > 1 ? VT1732_PDMAx_2BUFS : 0) |
		      get_format_bit(runtime->format) |
		      VT1732_PDMAx_FMT_STEREO);
	vt1732_writew(chip, base_reg + VT1732_PDMAx_TERM_COUNT,
		      (snd_pcm_lib_period_bytes(substream) >> 2) - 1);
	return 0;
}

static snd_pcm_uframes_t
stereo_pcm_pointer(struct snd_pcm_substream *substream, unsigned int nstr)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_vt1732_stream *str;
	unsigned int pos, idx, count;

	if (nstr == 0)
		return vt1732_xdma0_pointer(substream);
	str = &chip->str[substream->stream][nstr];
	pos = vt1732_readw(chip, str->base_reg + VT1732_PDMAx_BASE_COUNT);
	idx = (pos >> 16) & 1;
	pos = (pos & 0xffff) << 2;
	if (idx < str->nbufs - 1)
		count = MAX_DMA_BUF_SIZE;
	else
		count = str->buffer_bytes % MAX_DMA_BUF_SIZE;
	if (pos >= count) {
		printk(KERN_ERR PFX "invalid PDMAx count: 0x%x\n", pos);
		return 0;
	}
	pos = MAX_DMA_BUF_SIZE * idx + (count - 1 - pos);
	return bytes_to_frames(substream->runtime, pos);
}

static int stereo_playback_prepare(struct snd_pcm_substream *substream)
{
	return stereo_pcm_prepare(substream, substream->number * 2);
}

static int stereo_playback_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_playback_trigger(substream, cmd, substream->number * 2);
}

static snd_pcm_uframes_t
stereo_playback_pointer(struct snd_pcm_substream *substream)
{
	return stereo_pcm_pointer(substream, substream->number * 2);
}

static int stereo_capture_prepare(struct snd_pcm_substream *substream)
{
	return stereo_pcm_prepare(substream, substream->number * 2);
}

static int stereo_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_capture_trigger(substream, cmd, substream->number * 2);
}

static snd_pcm_uframes_t
stereo_capture_pointer(struct snd_pcm_substream *substream)
{
	return stereo_pcm_pointer(substream, substream->number * 2);
}

static int stereo_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned int i, nstr = substream->number * 2;

	mutex_lock(&chip->open_mutex);
	runtime->hw = stereo_pcm_hw;
	for (i = nstr; i < nstr + 1; i++) {
		if (vt1732_check_stream(chip, substream, i)) {
			mutex_unlock(&chip->open_mutex);
			return -EBUSY;
		}
	}
	vt1732_reserve_streams(chip, substream, nstr, nstr + 1);
	limit_rate(chip, runtime, 0);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   VT1732_BUFFER_ALIGN);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   VT1732_BUFFER_ALIGN);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static int stereo_pcm_close(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);

	mutex_lock(&chip->open_mutex);
	vt1732_release_streams(chip, substream);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static struct snd_pcm_ops stereo_playback_ops = {
	.open =		stereo_pcm_open,
	.close =	stereo_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	stereo_playback_prepare,
	.trigger =	stereo_playback_trigger,
	.pointer =	stereo_playback_pointer,
};

static struct snd_pcm_ops stereo_capture_ops = {
	.open =		stereo_pcm_open,
	.close =	stereo_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	stereo_capture_prepare,
	.trigger =	stereo_capture_trigger,
	.pointer =	stereo_capture_pointer,
};

/*
 * mono streams
 */

static struct snd_pcm_hardware mono_pcm_hw = {
	.info =		(SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED |
			 SNDRV_PCM_INFO_BLOCK_TRANSFER |
			 SNDRV_PCM_INFO_MMAP_VALID |
			 SNDRV_PCM_INFO_PAUSE),
	.formats =	(SNDRV_PCM_FMTBIT_U8 |
			 SNDRV_PCM_FMTBIT_S16_LE |
			 SNDRV_PCM_FMTBIT_S32_LE),
	.rates =	SNDRV_PCM_RATE_KNOT | SNDRV_PCM_RATE_8000_192000,
	.rate_min =	8000,
	.rate_max =	192000,
	.channels_min =		1,
	.channels_max =		1,
	.buffer_bytes_max =	256*1024,
	.period_bytes_min =	4 * 2,
	.period_bytes_max =	256*1024,
	.periods_min =		2,
	.periods_max =		1024,
};

static int mono_pcm_prepare(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct snd_vt1732_stream *str;
	unsigned int base_reg;
	unsigned int nstr = substream->number;
	int err;

	if (nstr == 0)
		return vt1732_xdma0_prepare(substream);

	err = snd_vt1732_set_rate(chip, runtime->rate, 0);
	if (err < 0)
		return err;

	str = &chip->str[substream->stream][substream->number];
	base_reg = str->base_reg;

	str->buffer_bytes = snd_pcm_lib_buffer_bytes(substream);
	str->nbufs = 1;

	vt1732_writel(chip, base_reg + VT1732_PDMAx_BASE_ADDR, 
		      runtime->dma_addr);
	vt1732_writew(chip, base_reg + VT1732_PDMAx_BASE_COUNT,
		      (str->buffer_bytes >> 2) - 1);
	vt1732_writeb(chip, base_reg + VT1732_PDMAx_FORMAT,
		      get_format_bit(runtime->format));
	vt1732_writew(chip, base_reg + VT1732_PDMAx_TERM_COUNT,
		      (snd_pcm_lib_period_bytes(substream) >> 2) - 1);
	return 0;
}

static snd_pcm_uframes_t
mono_pcm_pointer(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_vt1732_stream *str;
	unsigned int nstr = substream->number;
	unsigned int pos;

	if (nstr == 0)
		return vt1732_xdma0_pointer(substream);
	str = &chip->str[substream->stream][nstr];
	pos = vt1732_readw(chip, str->base_reg + VT1732_PDMAx_BASE_COUNT);
	pos = (pos & 0xffff) << 2;
	if (pos >= str->buffer_bytes) {
		printk(KERN_ERR PFX "invalid PDMAx mono count: 0x%x\n", pos);
		return 0;
	}
	pos = str->buffer_bytes - 1 - pos;
	return bytes_to_frames(substream->runtime, pos);
}

static int mono_playback_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_playback_trigger(substream, cmd, substream->number);
}

static int mono_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_capture_trigger(substream, cmd, substream->number);
}

static int mono_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	unsigned int nstr = substream->number;

	mutex_lock(&chip->open_mutex);
	runtime->hw = mono_pcm_hw;
	if (vt1732_check_stream(chip, substream, nstr)) {
		mutex_unlock(&chip->open_mutex);
		return -EBUSY;
	}
	vt1732_reserve_streams(chip, substream, nstr, nstr);
	limit_rate(chip, runtime, 0);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   VT1732_BUFFER_ALIGN);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   VT1732_BUFFER_ALIGN);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static int mono_pcm_close(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);

	mutex_lock(&chip->open_mutex);
	vt1732_release_streams(chip, substream);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static struct snd_pcm_ops mono_playback_ops = {
	.open =		mono_pcm_open,
	.close =	mono_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	mono_pcm_prepare,
	.trigger =	mono_playback_trigger,
	.pointer =	mono_pcm_pointer,
};

static struct snd_pcm_ops mono_capture_ops = {
	.open =		mono_pcm_open,
	.close =	mono_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	mono_pcm_prepare,
	.trigger =	mono_capture_trigger,
	.pointer =	mono_pcm_pointer,
};


/*
 * SPIDF I/O
 */

static struct snd_pcm_hardware spdif_pcm_hw = {
	.info =		(SNDRV_PCM_INFO_MMAP | SNDRV_PCM_INFO_INTERLEAVED |
			 SNDRV_PCM_INFO_BLOCK_TRANSFER |
			 SNDRV_PCM_INFO_MMAP_VALID |
			 SNDRV_PCM_INFO_PAUSE),
	.formats =	(SNDRV_PCM_FMTBIT_S16_LE |
			 SNDRV_PCM_FMTBIT_S32_LE),
	.rates =	(SNDRV_PCM_RATE_32000 |
			 SNDRV_PCM_RATE_44100 |
			 SNDRV_PCM_RATE_48000 |
			 SNDRV_PCM_RATE_88200 |
			 SNDRV_PCM_RATE_96000 |
			 SNDRV_PCM_RATE_176400 |
			 SNDRV_PCM_RATE_192000),
	.rate_min =	32000,
	.rate_max =	192000,
	.channels_min =		2,
	.channels_max =		2,
	.buffer_bytes_max =	(256*2*1024),
	.period_bytes_min =	2 * 4 * 2,
	.period_bytes_max =	256*1024,
	.periods_min =		2,
	.periods_max =		1024,
};

static int spdif_pcm_prepare(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		update_spdif_rate(chip, substream->runtime->rate);
	return stereo_pcm_prepare(substream, VT1732_PCM_SPDIF);
}

static int spdif_playback_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_playback_trigger(substream, cmd, VT1732_PCM_SPDIF);
}

static int spdif_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_capture_trigger(substream, cmd, VT1732_PCM_SPDIF);
}

static snd_pcm_uframes_t
spdif_pcm_pointer(struct snd_pcm_substream *substream)
{
	return stereo_pcm_pointer(substream, VT1732_PCM_SPDIF);
}

static int spdif_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	mutex_lock(&chip->open_mutex);
	runtime->hw = spdif_pcm_hw;
	if (limit_rate(chip, runtime, 1) < 0) {
		mutex_unlock(&chip->open_mutex);
		return -EBUSY;
	}
	vt1732_reserve_streams(chip, substream,
			       VT1732_PCM_SPDIF, VT1732_PCM_SPDIF);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   VT1732_BUFFER_ALIGN);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   VT1732_BUFFER_ALIGN);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static int spdif_pcm_close(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);

	mutex_lock(&chip->open_mutex);
	vt1732_release_streams(chip, substream);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static struct snd_pcm_ops spdif_playback_ops = {
	.open =		spdif_pcm_open,
	.close =	spdif_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	spdif_pcm_prepare,
	.trigger =	spdif_playback_trigger,
	.pointer =	spdif_pcm_pointer,
};

static struct snd_pcm_ops spdif_capture_ops = {
	.open =		spdif_pcm_open,
	.close =	spdif_pcm_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	spdif_pcm_prepare,
	.trigger =	spdif_capture_trigger,
	.pointer =	spdif_pcm_pointer,
};

/*
 * mixer loopback
 */

static int mixer_capture_prepare(struct snd_pcm_substream *substream)
{
	return stereo_pcm_prepare(substream, VT1732_PCM_MIXER);
}

static int mixer_capture_trigger(struct snd_pcm_substream *substream, int cmd)
{
	return vt1732_capture_trigger(substream, cmd, VT1732_PCM_MIXER);
}

static snd_pcm_uframes_t
mixer_capture_pointer(struct snd_pcm_substream *substream)
{
	return stereo_pcm_pointer(substream, VT1732_PCM_MIXER);
}

static int mixer_capture_open(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	mutex_lock(&chip->open_mutex);
	runtime->hw = spdif_pcm_hw;
	if (limit_rate(chip, runtime, 0) < 0) {
		mutex_unlock(&chip->open_mutex);
		return -EBUSY;
	}
	vt1732_reserve_streams(chip, substream,
			       VT1732_PCM_MIXER, VT1732_PCM_MIXER);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES,
				   VT1732_BUFFER_ALIGN);
	snd_pcm_hw_constraint_step(runtime, 0, SNDRV_PCM_HW_PARAM_BUFFER_BYTES,
				   VT1732_BUFFER_ALIGN);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static int mixer_capture_close(struct snd_pcm_substream *substream)
{
	struct snd_vt1732 *chip = snd_pcm_substream_chip(substream);

	mutex_lock(&chip->open_mutex);
	vt1732_release_streams(chip, substream);
	mutex_unlock(&chip->open_mutex);
	return 0;
}

static struct snd_pcm_ops mixer_capture_ops = {
	.open =		mixer_capture_open,
	.close =	mixer_capture_close,
	.ioctl =	snd_pcm_lib_ioctl,
	.hw_params =	vt1732_pcm_hw_params,
	.hw_free =	vt1732_pcm_hw_free,
	.prepare =	mixer_capture_prepare,
	.trigger =	mixer_capture_trigger,
	.pointer =	mixer_capture_pointer,
};


/*
 * Mixer stuff
 */

/*
 *  rate
 */
static int get_spdif_sync_index(struct snd_vt1732 *chip)
{
	switch (chip->rate_max) {
	case 192000:	return NUM_RATES_192;
	case 96000:	return NUM_RATES_96;
	default:	return NUM_RATES_48;
	}
}

static int vt1732_rate_lock_info(struct snd_kcontrol *kcontrol,
				 struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int vt1732_rate_lock_get(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);

	return chip->rate_locked;
}

static int vt1732_rate_lock_put(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	int locked = !! ucontrol->value.integer.value[0];
	int change;

	spin_lock_irq(&chip->reg_lock);
	change = (chip->rate_locked != locked);
	chip->rate_locked = locked;
	spin_unlock_irq(&chip->reg_lock);
	return change;
}

static int vt1732_internal_clock_info(struct snd_kcontrol *kcontrol,
				      struct snd_ctl_elem_info *uinfo)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	int spdif_sync = get_spdif_sync_index(chip);

	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	uinfo->value.enumerated.items = spdif_sync + 1;
	if (uinfo->value.enumerated.item >= uinfo->value.enumerated.items)
		uinfo->value.enumerated.item = uinfo->value.enumerated.items-1;
	if (uinfo->value.enumerated.item == spdif_sync)
		strcpy(uinfo->value.enumerated.name, "IEC958 Input");
	else
		strcpy(uinfo->value.enumerated.name,
		       rate_texts[uinfo->value.enumerated.item]);
	return 0;
}

static int vt1732_internal_clock_get(struct snd_kcontrol *kcontrol,
				     struct snd_ctl_elem_value *ucontrol)
{
	static unsigned char rate_bits_idx[16] = {
		9, 6, 3, 1, 7, 4, 0, 12, 8, 5, 2, 11, 13, 255, 14, 10
	};
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned char val;
	
	spin_lock_irq(&chip->reg_lock);
	if (chip->spdif_in_sync)
		ucontrol->value.enumerated.item[0] = get_spdif_sync_index(chip);
	else {
		val = vt1732_readb(chip, VT1732_RATE_SELECT) & 0x0f;
		val = rate_bits_idx[val];
		if (val == 255) {
			snd_BUG();
			val = 0;
		}
		ucontrol->value.enumerated.item[0] = val;
	}
	spin_unlock_irq(&chip->reg_lock);
	return 0;
}

static int vt1732_internal_clock_put(struct snd_kcontrol *kcontrol,
				     struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	int spdif_sync = get_spdif_sync_index(chip);
	int rate, force;

	if (ucontrol->value.enumerated.item[0] > spdif_sync)
		return -EINVAL;

	if (ucontrol->value.enumerated.item[0] == spdif_sync) {
		if (chip->spdif_in_sync)
			return 0;
		chip->spdif_in_sync = 1;
		return snd_vt1732_set_rate(chip, chip->rate, 1);
	}
	force = 0;
	if (chip->spdif_in_sync) {
		force = 1;
		chip->spdif_in_sync = 0;
	}
	rate = rates[ucontrol->value.enumerated.item[0]];
	return snd_vt1732_set_rate(chip, rate, force);
}

static struct snd_kcontrol_new vt1732_clock_ctls[] __devinitdata = {
	{
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.name = "Internal Clock",
		.info = vt1732_internal_clock_info,
		.get = vt1732_internal_clock_get,
		.put = vt1732_internal_clock_put
	},
	{
		.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
		.name = "Rate Locking",
		.info = vt1732_rate_lock_info,
		.get = vt1732_rate_lock_get,
		.put = vt1732_rate_lock_put
	},
};

static unsigned int encode_spdif_bits(struct snd_aes_iec958 *diga)
{
	unsigned int val, rbits;

	val = diga->status[0] & 0x03; /* professional, non-audio */
	if (val & 0x01) {
		/* professional */
		if ((diga->status[0] & IEC958_AES0_PRO_EMPHASIS) ==
		    IEC958_AES0_PRO_EMPHASIS_5015)
			val |= VT1732_SPDF_TX_PREEMP;
		rbits = (diga->status[4] >> 3) & 0x0f;
		if (rbits) {
			switch (rbits) {
			case 2: val |= VT1732_SPDF_TX_RATE_96; break;
			case 3: val |= VT1732_SPDF_TX_RATE_192; break;
			case 10: val |= VT1732_SPDF_TX_RATE_88; break;
			case 11: val |= VT1732_SPDF_TX_RATE_176; break;
			}
		} else {
			switch (diga->status[0] & IEC958_AES0_PRO_FS) {
			case IEC958_AES0_PRO_FS_44100:
				val |= VT1732_SPDF_TX_RATE_44;
				break;
			case IEC958_AES0_PRO_FS_32000:
				val |= VT1732_SPDF_TX_RATE_32;
				break;
			default:
				val |= VT1732_SPDF_TX_RATE_48;
				break;
			}
		}
	} else {
		/* consumer */
		val |= diga->status[1] & VT1732_SPDF_TX_COPY; /* copyright */
		if ((diga->status[0] & IEC958_AES0_CON_EMPHASIS) ==
		    IEC958_AES0_CON_EMPHASIS_5015)
			val |= VT1732_SPDF_TX_PREEMP;
		/* category */
		val |= (unsigned int)(diga->status[1] & 0x3f) <<
			VT1732_SPDF_TX_CATEGORY_SHIFT;
		val |= (unsigned int)(diga->status[3] & IEC958_AES3_CON_FS)
			<< VT1732_SPDF_TX_RATE_SHIFT;
	}
	return val;
}

static void decode_spdif_bits(struct snd_aes_iec958 *diga, unsigned int val)
{
	memset(diga->status, 0, sizeof(diga->status));
	diga->status[0] = val & 0x03; /* professional, non-audio */
	if (val & VT1732_SPDF_TX_PROF) {
		/* professional */
		/* emphasis */
		if (val & VT1732_SPDF_TX_PREEMP)
			diga->status[0] |= IEC958_AES0_PRO_EMPHASIS_5015;
		/* fs */
		switch (val & VT1732_SPDF_TX_RATE) {
		case VT1732_SPDF_TX_RATE_44:
			break;
		case VT1732_SPDF_TX_RATE_32:
			diga->status[0] |= IEC958_AES0_PRO_FS_32000;
			break;
		case VT1732_SPDF_TX_RATE_48:
		default:
			diga->status[0] |= IEC958_AES0_PRO_FS_48000;
			break;
		case VT1732_SPDF_TX_RATE_96:
			diga->status[4] |= 2 << 3;
			break;
		case VT1732_SPDF_TX_RATE_192:
			diga->status[4] |= 3 << 3;
			break;
		case VT1732_SPDF_TX_RATE_88:
			diga->status[4] |= 10 << 3;
			break;
		case VT1732_SPDF_TX_RATE_176:
			diga->status[4] |= 11 << 3;
			break;
		}
	} else {
		/* consumer */
		/* copyright */
		diga->status[0] |= val & (VT1732_SPDF_TX_COPY);
		/* emphasis */
		if (val & VT1732_SPDF_TX_PREEMP)
			diga->status[0] |= IEC958_AES0_CON_EMPHASIS_5015;
		/* category */
		diga->status[1] |= (val & VT1732_SPDF_TX_CATEGORY)
			>> VT1732_SPDF_TX_CATEGORY_SHIFT;
		/* fs */
		diga->status[3] |= (val & VT1732_SPDF_TX_RATE)
			>> VT1732_SPDF_TX_RATE_SHIFT;
	}
}

static int vt1732_spdif_default_info(struct snd_kcontrol *kcontrol,
				     struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_IEC958;
	uinfo->count = 1;
	return 0;
}

static int vt1732_spdif_default_get(struct snd_kcontrol *kcontrol,
				    struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int val;

	val = vt1732_readw(chip, VT1732_SPDF_TX_CTRL);
	decode_spdif_bits(&ucontrol->value.iec958, val);
	return 0;
}

static int vt1732_spdif_default_put(struct snd_kcontrol *kcontrol,
				    struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int val, old;

	val = encode_spdif_bits(&ucontrol->value.iec958);
	spin_lock_irq(&chip->reg_lock);
	old = vt1732_readw(chip, VT1732_SPDF_TX_CTRL);
	if (val != old)
		update_spdif_bits(chip, val);
	spin_unlock_irq(&chip->reg_lock);
	return (val != old);
}

static int vt1732_spdif_maskc_get(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.iec958.status[0] = IEC958_AES0_NONAUDIO |
		IEC958_AES0_PROFESSIONAL |
		IEC958_AES0_CON_NOT_COPYRIGHT |
		IEC958_AES0_CON_EMPHASIS;
	ucontrol->value.iec958.status[1] = IEC958_AES1_CON_ORIGINAL |
		IEC958_AES1_CON_CATEGORY;
	ucontrol->value.iec958.status[3] = IEC958_AES3_CON_FS;
	return 0;
}

static int vt1732_spdif_maskp_get(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.iec958.status[0] = IEC958_AES0_NONAUDIO |
		IEC958_AES0_PROFESSIONAL |
		IEC958_AES0_PRO_FS |
		IEC958_AES0_PRO_EMPHASIS;
	return 0;
}

static int vt1732_spdif_sw_info(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int vt1732_spdif_sw_get(struct snd_kcontrol *kcontrol,
			       struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);

	if (vt1732_readb(chip, VT1732_SPDIF_CONFIG) & VT1732_CFG_SPDF_EN_TX)
		ucontrol->value.integer.value[0] = 1;
	else
		ucontrol->value.integer.value[0] = 0;
	return 0;
}

static int vt1732_spdif_sw_put(struct snd_kcontrol *kcontrol,
			       struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned char old, val;

	spin_lock_irq(&chip->reg_lock);
	old = val = vt1732_readb(chip, VT1732_SPDIF_CONFIG);
	val &= ~VT1732_CFG_SPDF_EN_TX;
	if (ucontrol->value.integer.value[0])
		val |= VT1732_CFG_SPDF_EN_TX;
	if (old != val)
		vt1732_writeb(chip, VT1732_SPDIF_CONFIG, val);
	spin_unlock_irq(&chip->reg_lock);
	return old != val;
}

static struct snd_kcontrol_new vt1732_spdif_ctls[] __devinitdata =
{
	{
		.iface =	SNDRV_CTL_ELEM_IFACE_PCM,
		.name =         SNDRV_CTL_NAME_IEC958("",PLAYBACK,DEFAULT),
		.info =		vt1732_spdif_default_info,
		.get =		vt1732_spdif_default_get,
		.put =		vt1732_spdif_default_put
	},
	{
		.access =	SNDRV_CTL_ELEM_ACCESS_READ,
		.iface =	SNDRV_CTL_ELEM_IFACE_PCM,
		.name =         SNDRV_CTL_NAME_IEC958("",PLAYBACK,CON_MASK),
		.info =		vt1732_spdif_default_info,
		.get =		vt1732_spdif_maskc_get,
	},
	{
		.access =	SNDRV_CTL_ELEM_ACCESS_READ,
		.iface =	SNDRV_CTL_ELEM_IFACE_PCM,
		.name =         SNDRV_CTL_NAME_IEC958("",PLAYBACK,PRO_MASK),
		.info =		vt1732_spdif_default_info,
		.get =		vt1732_spdif_maskp_get,
	},
	{
		.iface =	SNDRV_CTL_ELEM_IFACE_MIXER,
		/* FIXME: the following conflict with IEC958 Playback Route */
		/* .name =         SNDRV_CTL_NAME_IEC958("",PLAYBACK,SWITCH), */
		.name =         SNDRV_CTL_NAME_IEC958("Output ",NONE,SWITCH),
		.info =		vt1732_spdif_sw_info,
		.get =		vt1732_spdif_sw_get,
		.put =		vt1732_spdif_sw_put
	},
};

/* routing table */
static int vt1732_route_info(struct snd_kcontrol *kcontrol,
			     struct snd_ctl_elem_info *uinfo)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int nitems;

	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	nitems = 2; /* PCM & Monitor */
	nitems += chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE]; /* loopback */
	nitems += chip->spdif_in * 2; /* SPDIF-loopback L/R */

	uinfo->value.enumerated.items = nitems;
	if (uinfo->value.enumerated.item >= nitems)
		uinfo->value.enumerated.item = nitems - 1;
	if (uinfo->value.enumerated.item == 0)
		strcpy(uinfo->value.enumerated.name, "PCM");
	else if (uinfo->value.enumerated.item == 1)
		strcpy(uinfo->value.enumerated.name, "Monitor");
	else if (uinfo->value.enumerated.item <=
		 chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 1)
		sprintf(uinfo->value.enumerated.name, "Analog In %d",
			uinfo->value.enumerated.item - 2);
	else
		sprintf(uinfo->value.enumerated.name, "SPDIF In %c",
			(uinfo->value.enumerated.item ==
			 chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 2) ?
			'L' : 'R');
	return 0;
}

static unsigned int analog_route_shift[16] = {
	0, 8, 2, 10, 4, 12, 6, 14, 16, 24, 18, 26, 20, 28, 22, 30
};

static unsigned int get_analog_route_val(struct snd_vt1732 *chip,
					 unsigned int idx)
{
	unsigned int val;

	val = vt1732_readl(chip, VT1732_LOOPBACK_OUT);
	val = (val >> analog_route_shift[idx]) & 0x03;
	if (val <= 1)
		return val; /* PCM  or Monitor */
	else if (val == 2) {
		/* analog-in */
		val = vt1732_readb(chip, VT1732_PSDOUT0_0_SRC + idx);
		return (val & 0x0f) + 2;
	} else {
		/* SPDIF-in */
		val = vt1732_readb(chip, VT1732_PSDOUT0_0_SRC + idx);
		return chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 2 +
			((val >> 7) & 1);
	}
}

static void put_analog_route_val(struct snd_vt1732 *chip,
				 unsigned int idx, unsigned int val)
{
	unsigned int nval, oval;

	if (val <= 1)
		nval = val;
	else if (val <= chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 1)
		nval = 2;
	else
		nval = 3;

	oval = vt1732_readl(chip, VT1732_LOOPBACK_OUT);
	oval &= ~(3 << analog_route_shift[idx]);
	oval |= (nval << analog_route_shift[idx]);
	vt1732_writel(chip, VT1732_LOOPBACK_OUT, oval);
	if (val <= 1)
		return;

	oval = vt1732_readb(chip, VT1732_PSDOUT0_0_SRC + idx);
	if (nval == 2) {
		/* analog-in */
		oval &= ~0x0f;
		oval |= (val - 2);
	} else {
		/* SPDIF-in */
		oval &= ~(1 << 7);
		if (val == chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 2)
			oval |= (1 << 7);
	}
	vt1732_writeb(chip, VT1732_PSDOUT0_0_SRC + idx, oval);
}

static int vt1732_analog_route_get(struct snd_kcontrol *kcontrol,
				   struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id);

	ucontrol->value.enumerated.item[0] = get_analog_route_val(chip, idx);
	return 0;
}

static int vt1732_analog_route_put(struct snd_kcontrol *kcontrol,
				   struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id);
	unsigned int val;
	int change = 0;

	spin_lock_irq(&chip->reg_lock);
	val = get_analog_route_val(chip, idx);
	if (val != ucontrol->value.enumerated.item[0]) {
		put_analog_route_val(chip, idx,
				     ucontrol->value.enumerated.item[0]);
		change = 1;
	}
	spin_unlock_irq(&chip->reg_lock);
	return change;
}

static unsigned int get_spdif_route_val(struct snd_vt1732 *chip,
					unsigned int idx)
{
	unsigned int val, oval;

	oval = vt1732_readl(chip, VT1732_LOOPBACK_SPDIF);
	val = (oval >> (idx << 1)) & 0x03;
	if (val <= 1)
		return val; /* PCM  or Monitor */
	else if (val == 2) {
		/* analog-in */
		val = (oval >> (8 + idx * 4)) & 0x07;
		val |= (oval >> (idx + 1)) & 0x08;
		return val + 2;
	} else {
		/* SPDIF-in */
		val = (oval >> (11 + idx * 4)) & 1;
		return chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 2 + val;
	}
}

static void put_spdif_route_val(struct snd_vt1732 *chip,
				unsigned int idx, unsigned int val)
{
	unsigned int nval, oval;

	if (val <= 1)
		nval = val;
	else if (val <= chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 1)
		nval = 2;
	else
		nval = 3;

	oval = vt1732_readl(chip, VT1732_LOOPBACK_SPDIF);
	oval &= ~(3 << (idx << 1));
	oval |= (nval << (idx << 1));
	if (nval == 2) {
		/* analog-in */
		val -= 2;
		oval &= ~(1 << (idx + 4));
		oval |= (val & 0x08) << (idx + 1);
		oval &= ~(7 << (8 + idx * 4));
		oval |= (val & 0x07) << (8 + idx * 4);
	} else {
		/* SPDIF-in */
		val -= chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] + 2;
		oval &= ~(1 << (11 + idx * 4));
		oval |= (val << (11 + idx * 4));
	}
	vt1732_writeb(chip, VT1732_LOOPBACK_SPDIF, oval);
}

static int vt1732_spdif_route_get(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id);

	ucontrol->value.enumerated.item[0] = get_spdif_route_val(chip, idx);
	return 0;
}

static int vt1732_spdif_route_put(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id);
	unsigned int val;
	int change = 0;

	spin_lock_irq(&chip->reg_lock);
	val = get_spdif_route_val(chip, idx);
	if (val != ucontrol->value.enumerated.item[0]) {
		put_spdif_route_val(chip, idx,
				    ucontrol->value.enumerated.item[0]);
		change = 1;
	}
	spin_unlock_irq(&chip->reg_lock);
	return change;
}

static struct snd_kcontrol_new vt1732_analog_route_ctl __devinitdata = {
	.iface = SNDRV_CTL_ELEM_IFACE_PCM,
	.name = "Analog Playback Route",
	.info = vt1732_route_info,
	.get = vt1732_analog_route_get,
	.put = vt1732_analog_route_put,
	/* count is set later */
};

static struct snd_kcontrol_new vt1732_spdif_route_ctl __devinitdata = {
	.iface = SNDRV_CTL_ELEM_IFACE_PCM,
	.name = SNDRV_CTL_NAME_IEC958("",PLAYBACK,NONE) "Route",
	.info = vt1732_route_info,
	.get = vt1732_spdif_route_get,
	.put = vt1732_spdif_route_put,
	.count = 2,
};

/* input monitoring mixer */
static void update_monitor_mixer(struct snd_vt1732 *chip, unsigned int idx)
{
	static unsigned char vol_idx[MONITOR_VOLS] = {
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, /* PDMA0-7 */
		0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, /* PDMA8-15 */
		0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, /* RDMA0-7 */
		0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, /* RMDA8-15 */
		0x08, 0x09, /* PSP0-1 */
		0x12, 0x13, /* RSP0-1 */
		0x14, /* MIX Master */
	};
	unsigned short val;

	val = chip->monitor_vol[idx];
	if (!(val & 0x80))
		val |= 0x00ff; /* mute left */
	if (!(val & 0x8000))
		val |= 0xff00; /* mute right */

	vt1732_writeb(chip, VT1732_VOL_CTRL_STREAM, vol_idx[idx]);
	vt1732_writew(chip, VT1732_VOL_CTRL_DATA, val);
}

static int vt1732_monitor_sw_info(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_BOOLEAN;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 1;
	return 0;
}

static int vt1732_monitor_sw_get(struct snd_kcontrol *kcontrol,
				 struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id) +
		kcontrol->private_value;

	spin_lock_irq(&chip->reg_lock);
	ucontrol->value.integer.value[0] = (chip->monitor_vol[idx] >> 7) & 1;
	ucontrol->value.integer.value[1] = (chip->monitor_vol[idx] >> 15) & 1;
	spin_unlock_irq(&chip->reg_lock);
	return 0;
}

static int vt1732_monitor_sw_put(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id) +
		kcontrol->private_value;
	unsigned int nval;
	int change = 0;

	nval = (ucontrol->value.integer.value[0] ? 0x80 : 0) |
		(ucontrol->value.integer.value[1] ? (0x80 << 8) : 0);
	spin_lock_irq(&chip->reg_lock);
	nval |= chip->monitor_vol[idx] & ~0x8080;
	if (nval != chip->monitor_vol[idx]) {
		change = 1;
		chip->monitor_vol[idx] = nval;
		update_monitor_mixer(chip, idx);
	}
	spin_unlock_irq(&chip->reg_lock);
	return change;
}

static int vt1732_monitor_vol_info(struct snd_kcontrol *kcontrol,
				   struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 96;
	return 0;
}

static int vt1732_monitor_vol_get(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id) +
		kcontrol->private_value;

	spin_lock_irq(&chip->reg_lock);
	ucontrol->value.integer.value[0] = chip->monitor_vol[idx] & 0x7f;
	ucontrol->value.integer.value[1] = (chip->monitor_vol[idx] >> 8) & 0x7f;
	spin_unlock_irq(&chip->reg_lock);
	return 0;
}

static int vt1732_monitor_vol_put(struct snd_kcontrol *kcontrol,
				  struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id) +
		kcontrol->private_value;
	unsigned int nval;
	int change = 0;

	nval = (ucontrol->value.integer.value[0] & 0x7f) |
		((ucontrol->value.integer.value[1] & 0x7f) << 8);
	spin_lock_irq(&chip->reg_lock);
	nval |= chip->monitor_vol[idx] & ~0x7f7f;
	if (nval != chip->monitor_vol[idx]) {
		change = 1;
		chip->monitor_vol[idx] = nval;
		update_monitor_mixer(chip, idx);
	}
	spin_unlock_irq(&chip->reg_lock);
	return change;
}

static int __devinit new_monitor_ctls(struct snd_vt1732 *chip, const char *pfx,
				      unsigned int count, unsigned int type_ofs)
{
	static const char *sfx[] = { "Switch", "Volume" };
	static struct snd_kcontrol_new monitor_ctls[] = {
		{
			.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
			.info = vt1732_monitor_sw_info,
			.get = vt1732_monitor_sw_get,
			.put = vt1732_monitor_sw_put,
		},
		{
			.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
			.info = vt1732_monitor_vol_info,
			.get = vt1732_monitor_vol_get,
			.put = vt1732_monitor_vol_put,
		},
	};
	struct snd_kcontrol *kctl;
	int i, err;

	for (i = 0; i < 2; i++) {
		char tmp[32];
		monitor_ctls[i].count = count;
		monitor_ctls[i].private_value = type_ofs;
		snprintf(tmp, sizeof(tmp), "%s %s", pfx, sfx[i]);
		monitor_ctls[i].name = tmp;
		kctl = snd_ctl_new1(&monitor_ctls[i], chip);
		if (!kctl)
			return -ENOMEM;
		err = snd_ctl_add(chip->card, kctl);
		if (err < 0)
			return err;
	}
	return 0;
}

/* vu meter */
static int vt1732_vol_meter_info(struct snd_kcontrol *kcontrol,
				 struct snd_ctl_elem_info *uinfo)
{
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 1;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 255;
	return 0;
}

static int vt1732_vol_meter_get(struct snd_kcontrol *kcontrol,
				struct snd_ctl_elem_value *ucontrol)
{
	struct snd_vt1732 *chip = snd_kcontrol_chip(kcontrol);
	unsigned int idx = snd_ctl_get_ioffidx(kcontrol, &ucontrol->id) +
		kcontrol->private_value;
	unsigned int val;
	
	spin_lock_irq(&chip->reg_lock);
	vt1732_writeb(chip, VT1732_VOL_METER_INDEX, idx);
	val = vt1732_readl(chip, VT1732_VOL_METER_DATA);
	spin_unlock_irq(&chip->reg_lock);

	ucontrol->value.integer.value[0] = (val >> ((idx & 3) << 3)) & 0xff;
	return 0;
}

static int __devinit new_vol_meter_ctls(struct snd_vt1732 *chip,
					char *name, unsigned int count,
					unsigned int type_ofs)
{
	static struct snd_kcontrol_new vol_meter_ctl = {
		.iface = SNDRV_CTL_ELEM_IFACE_PCM,
		.access = (SNDRV_CTL_ELEM_ACCESS_READ |
			   SNDRV_CTL_ELEM_ACCESS_VOLATILE),
		.info = vt1732_vol_meter_info,
		.get = vt1732_vol_meter_get,
	};

	vol_meter_ctl.name = name;
	vol_meter_ctl.count = count;
	vol_meter_ctl.private_value = type_ofs;
	return snd_ctl_add(chip->card, snd_ctl_new1(&vol_meter_ctl, chip));
}


static int __devinit build_controls(struct snd_vt1732 *chip)
{
	int i, err;

	/* clock controls */
	for (i = 0; i < ARRAY_SIZE(vt1732_clock_ctls); i++) {
		err = snd_ctl_add(chip->card,
				  snd_ctl_new1(&vt1732_clock_ctls[i], chip));
		if (err < 0)
			return err;
	}

	/* built-in SPDIF output controls */
	if (chip->eeprom.data[EEPROM_SPDIF] & VT1732_CFG_SPDF_INT_TX) {
		for (i = 0; i < ARRAY_SIZE(vt1732_spdif_ctls); i++) {
			err = snd_ctl_add(chip->card,
					  snd_ctl_new1(&vt1732_spdif_ctls[i],
						       chip));
			if (err < 0)
				return err;
		}
	}

	/* analog routing */
	if (chip->hw_nstreams[SNDRV_PCM_STREAM_PLAYBACK]) {
		vt1732_analog_route_ctl.count = 
			chip->hw_nstreams[SNDRV_PCM_STREAM_PLAYBACK];
		err = snd_ctl_add(chip->card,
				  snd_ctl_new1(&vt1732_analog_route_ctl, chip));
		if (err < 0)
			return err;
	}
	/* spdif routing */
	if (chip->spdif_out) {
		err = snd_ctl_add(chip->card,
				  snd_ctl_new1(&vt1732_spdif_route_ctl, chip));
		if (err < 0)
			return err;
	}

	/* input monitor volumes/switches */
	err = new_monitor_ctls(chip, "Monitor Master Playback", 1,
			       MONITOR_MASTER);
	if (err < 0)
		return err;
	if (chip->hw_nstreams[0]) {
		err = new_monitor_ctls(chip, "Analog Output Monitor",
				       chip->hw_nstreams[0],
				       MONITOR_AOUT_0);
		if (err < 0)
			return err;
	}
	if (chip->spdif_out) {
		err = new_monitor_ctls(chip, "SPDIF Output Monitor", 2,
				       MONITOR_SPOUT_0);
		if (err < 0)
			return err;
	}
	if (chip->hw_nstreams[1]) {
		err = new_monitor_ctls(chip, "Analog Input Monitor",
				       chip->hw_nstreams[1],
				       MONITOR_AIN_0);
		if (err < 0)
			return err;
	}
	if (chip->spdif_in) {
		err = new_monitor_ctls(chip, "SPDIF Input Monitor", 2,
				       MONITOR_SPIN_0);
		if (err < 0)
			return err;
	}

	/* VU meters */
	err = new_vol_meter_ctls(chip, "Monitor Master Peak", 2,
				 VOLMETER_MASTER_0);
	if (err < 0)
		return err;
	if (chip->hw_nstreams[0]) {
		new_vol_meter_ctls(chip, "Analog Output Peak",
				   chip->hw_nstreams[0],
				   VOLMETER_MASTER_0);
		if (err < 0)
			return err;
	}
	if (chip->spdif_out) {
		err = new_vol_meter_ctls(chip, "SPDIF Output Peak", 2,
					 VOLMETER_SPOUT_0);
		if (err < 0)
			return err;
	}
	if (chip->hw_nstreams[1]) {
		err = new_vol_meter_ctls(chip, "Analog Input Peak",
					 chip->hw_nstreams[1],
					 VOLMETER_AIN_0);
		if (err< 0)
			return err;
	}
	if (chip->spdif_in) {
		err = new_vol_meter_ctls(chip, "SPDIF Input Peak", 2,
					 VOLMETER_SPIN_0);
		if (err < 0)
			return err;
	}

	/* board-specific controls */
	if (chip->config->build_controls) {
		err = chip->config->build_controls(chip);
		if (err < 0)
			return err;
	}

	return 0;
}

/*
 * MPU401
 */

#if 0
static void vt1732_dma_clear_bit(struct snd_vt1732 *chip, unsigned int bits)
{
	spin_lock_irq(&chip->reg_lock);
	vt1732_writew(chip, VT1732_DMA_INT_MASK,
		      vt1732_readw(chip, VT1732_DMA_INT_MASK) & ~bits);
	spin_unlock_irq(&chip->reg_lock);
}

static void vt1732_dma_set_bit(struct snd_vt1732 *chip, unsigned int bits)
{
	spin_lock_irq(&chip->reg_lock);
	vt1732_writew(chip, VT1732_DMA_INT_MASK,
		      vt1732_readw(chip, VT1732_DMA_INT_MASK) | bits);
	spin_unlock_irq(&chip->reg_lock);
}
#endif

static int __devinit vt1732_mpu401_new(struct snd_vt1732 *chip, int idx,
				       const char *name)
{
	int err;

	err = snd_mpu401_uart_new(chip->card, 0, MPU401_HW_ICE1712,
				  (long)(chip->iomap +
					 VT1732_UART1_DATA + idx * 2),
				  (MPU401_INFO_INTEGRATED | MPU401_INFO_MMIO |
				   MPU401_INFO_TX_IRQ),
				  chip->irq, 0,
				  &chip->rmidi[idx]);
	if (err < 0)
		return err;
	chip->uart[idx] = chip->rmidi[idx]->private_data;
	if (name)
		snprintf(chip->rmidi[idx]->name, sizeof(chip->rmidi[idx]->name),
			 "%s %d", name, chip->card->number);

	/* clear irq masks */
	spin_lock_irq(&chip->reg_lock);
	vt1732_writew(chip, VT1732_DMA_INT_MASK,
		      vt1732_readw(chip, VT1732_DMA_INT_MASK) &
		      ~((VT1732_UART1_RX_INT_MASK | VT1732_UART1_TX_INT_MASK)
			<< (1 - idx)));
	spin_unlock_irq(&chip->reg_lock);

	return 0;
}

static int __devinit create_midi(struct snd_vt1732 *chip)
{
	int err;

	if (!chip->has_midi) {
		err = vt1732_mpu401_new(chip, 0, chip->config->midi_name);
		if (err < 0)
			return err;
	}
	if (chip->has_midi2) {
		err = vt1732_mpu401_new(chip, 1, chip->config->midi2_name);
		if (err < 0)
			return err;
	}
	return 0;
}

/*
 * PCMs
 */

static int __devinit create_pcm_interleave(struct snd_vt1732 *chip, int device)
{
	struct snd_pcm *pcm;
	int err;

	err = snd_pcm_new(chip->card, "VT1732 Interleaved", device, 1, 1, &pcm);
	if (err < 0)
		return err;

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
			&interleave_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
			&interleave_capture_ops);

	pcm->private_data = chip;
	pcm->info_flags = 0;

	strcpy(pcm->name, "VT1732 Interleave");
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
					      snd_dma_pci_data(chip->pci),
					      1024*1024, 256*1024);
	chip->pcm_interleave = pcm;
	return 0;
}

static int __devinit create_pcm_stereo(struct snd_vt1732 *chip, int device)
{
	struct snd_pcm *pcm;
	int err;
	int nplay, ncapt;

	nplay = chip->hw_nstreams[SNDRV_PCM_STREAM_PLAYBACK] / 2;
	ncapt = chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE] / 2; 
	err = snd_pcm_new(chip->card, "VT1732 Stereo", device,
			  nplay, ncapt, &pcm);
	if (err < 0)
		return err;

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
			&stereo_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
			&stereo_capture_ops);

	pcm->private_data = chip;
	pcm->info_flags = 0;

	strcpy(pcm->name, "VT1732 Stereo");
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
					      snd_dma_pci_data(chip->pci),
					      256*1024, 0);
	chip->pcm_stereo = pcm;
	return 0;
}

static int __devinit create_pcm_mono(struct snd_vt1732 *chip, int device)
{
	struct snd_pcm *pcm;
	int err;
	int nplay, ncapt;

	nplay = chip->hw_nstreams[SNDRV_PCM_STREAM_PLAYBACK];
	ncapt = chip->hw_nstreams[SNDRV_PCM_STREAM_CAPTURE];
	err = snd_pcm_new(chip->card, "VT1732 Mono", device,
			  nplay, ncapt, &pcm);
	if (err < 0)
		return err;

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
			&mono_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
			&mono_capture_ops);

	pcm->private_data = chip;
	pcm->info_flags = 0;

	strcpy(pcm->name, "VT1732 Mono");
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
					      snd_dma_pci_data(chip->pci),
					      256*1024, 0);
	chip->pcm_mono = pcm;
	return 0;
}

static int __devinit create_pcm_spdif(struct snd_vt1732 *chip, int device)
{
	struct snd_pcm *pcm;
	int nplay, ncapt, err;

	nplay = chip->spdif_out ? 1 : 0;
	ncapt = chip->spdif_in ? 1 : 0;
	if (!nplay && !ncapt)
		return 0;

	err = snd_pcm_new(chip->card, "VT1732 SPDIF", device,
			  nplay, ncapt, &pcm);
	if (err < 0)
		return err;

	if (nplay)
		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
				&spdif_playback_ops);
	if (ncapt)
		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
				&spdif_capture_ops);

	pcm->private_data = chip;
	pcm->info_flags = 0;

	strcpy(pcm->name, "VT1732 SPDIF");
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
					      snd_dma_pci_data(chip->pci),
					      256*1024, 0);
	chip->pcm_spdif = pcm;
	return 0;
}

static int __devinit create_pcm_mixer(struct snd_vt1732 *chip, int device)
{
	struct snd_pcm *pcm;
	int err;

	err = snd_pcm_new(chip->card, "VT1732 Mixer", device, 0, 1, &pcm);
	if (err < 0)
		return err;

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
			&mixer_capture_ops);

	pcm->private_data = chip;
	pcm->info_flags = 0;

	strcpy(pcm->name, "VT1732 Mixer");
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
					      snd_dma_pci_data(chip->pci),
					      256*1024, 0);
	chip->pcm_mixer = pcm;
	return 0;
}


/*
 */

unsigned char __devinit snd_vt1732_read_i2c(struct snd_vt1732 *chip,
					    unsigned char id,
					    unsigned char dev,
					    unsigned char addr)
{
	long t = 0x10000;

	vt1732_writeb(chip, VT1732_I2C_ENABLE, (id << 4) | (1 << id));
	vt1732_writeb(chip, VT1732_I2C_BYTE_ADDR, addr);
	vt1732_writeb(chip, VT1732_I2C_DEV_ADDR, dev & ~VT1732_I2C_WRITE);
	while (t-- > 0 &&
	       (vt1732_readb(chip, VT1732_I2C_STATUS) & VT1732_I2C_BUSY))
		       ;
	return vt1732_readb(chip, VT1732_I2C_DATA);
}

/*
 */

/* do software reset and enable Envy24-II mode */
static void vt1732_reset(struct snd_vt1732 *chip)
{
	unsigned long timeout;

	vt1732_writeb(chip, VT1732_LEGACY_CTRL, VT1732_RESET);
	timeout = jiffies + HZ;
	do {
		schedule_timeout_uninterruptible(1);
		vt1732_writeb(chip, VT1732_LEGACY_CTRL, VT1732_NATIVE);
	} while (!(vt1732_readb(chip, VT1732_LEGACY_CTRL) & VT1732_NATIVE) &&
		 time_after_eq(timeout, jiffies));
}

static const struct snd_vt1732_card_info test_models[] = {
	{
		.subvendor = 0x14120000,
		.name = "Test Model",
		.model = "test",
		.eeprom = {
			(7 << 3) | (7 << 0),
			VT1732_CFG_I2S,
			(VT1732_CFG_I2S_96K|VT1732_CFG_I2S_FMT_24|
			 VT1732_CFG_I2S_192K),
			VT1732_CFG_SPDF_EN_TX|VT1732_CFG_SPDF_INT_TX
		},
	},
	{ }
};

static const struct snd_vt1732_card_info noboard_config;

static const struct snd_vt1732_card_info *card_tables[] __devinitdata = {
	test_models,
	NULL /* terminator */
};


static int __devinit look_for_card(struct snd_vt1732 *chip, const char *name)
{
	const struct snd_vt1732_card_info **tbl, *c;

	for (tbl = card_tables; *tbl; tbl++) {
		for (c = *tbl; c->subvendor; c++) {
			if (name && c->model && !strcmp(c->model, name)) {
				printk(KERN_INFO PFX "Using board model %s\n",
				       c->name);
				goto found;
			}
			if (c->subvendor == chip->eeprom.subvendor)
				goto found;
		}
	}
	return -ENODEV;

 found:
	chip->config = c;
	memcpy(chip->eeprom.data, c->eeprom, 4);
	if (c->driver)
		strcpy(chip->card->driver, c->driver);
	if (c->shortname)
		strcpy(chip->card->shortname, c->shortname);
	return 0;
}

static int __devinit read_eeprom(struct snd_vt1732 *chip)
{
	const int dev = 0xa0;		/* EEPROM device address */
	int i;

	chip->eeprom.subvendor = 0;
	if (!(vt1732_readb(chip, VT1732_I2C_STATUS) &
	       VT1732_I2C_EEPROM_PRESENT))
		return -ENODEV;

	for (i = 0; i < 4; i++)
		chip->eeprom.subvendor |=
			snd_vt1732_read_i2c(chip, 0, dev, i) << (i * 8);

	if (chip->eeprom.subvendor == 0 ||
	    chip->eeprom.subvendor == (unsigned int)-1) {
		/* invalid subvendor from EEPROM,
		 * try the PCI subststem ID instead
		 */
		u16 vendor, device;
		pci_read_config_word(chip->pci, PCI_SUBSYSTEM_VENDOR_ID,
				     &vendor);
		pci_read_config_word(chip->pci, PCI_SUBSYSTEM_ID,
				     &device);
		chip->eeprom.subvendor = ((unsigned int)swab16(vendor) << 16) |
			swab16(device);
		if (chip->eeprom.subvendor == 0 ||
		    chip->eeprom.subvendor == (unsigned int)-1)
			return -ENODEV;
	}

	if (look_for_card(chip, NULL) < 0) {
		printk(KERN_WARNING PFX "No matching model found for ID 0x%x\n",
		       chip->eeprom.subvendor);
		chip->config = &noboard_config;
		for (i = 0; i < 4; i++)
			chip->eeprom.data[i] =
				snd_vt1732_read_i2c(chip, 0, dev, i + 4);
	}

	return 0;
}

static int __devinit model_init(struct snd_vt1732 *chip, const char *modelname)
{
	if (!modelname) {
		if (read_eeprom(chip) < 0) {
			printk(KERN_ERR PFX
			       "No available configuration is found\n");
			return -ENODEV;
		}
	} else if (look_for_card(chip, modelname) < 0) {
		printk(KERN_ERR PFX "No board found for model %s\n", modelname);
		return -ENODEV;
	}

	if (!chip->config)
		return -ENODEV;

	if (!(chip->eeprom.data[EEPROM_IFACE] & VT1732_CFG_I2S)) {
		printk(KERN_ERR PFX "AC97 interface is not supported yet\n");
		return -ENODEV;
	}

	/* parse EEPROM data */
	chip->hw_nstreams[0] = ((chip->eeprom.data[EEPROM_CONFIG] & 7) + 1) * 2;
	chip->hw_nstreams[1] =
		(((chip->eeprom.data[EEPROM_CONFIG] >> 3) & 7) + 1) * 2;
	if (!chip->config->no_mpu401) {
		chip->has_midi = 1;
		if (chip->eeprom.data[EEPROM_CONFIG] & VT1732_CFG_UART2)
			chip->has_midi2 = 1;
	}

	if (chip->eeprom.data[EEPROM_I2S] & VT1732_CFG_I2S_192K)
		chip->rate_max = 192000;
	else if (chip->eeprom.data[EEPROM_I2S] & VT1732_CFG_I2S_96K)
		chip->rate_max = 96000;
	else
		chip->rate_max = 48000;

	if (chip->eeprom.data[EEPROM_SPDIF] &
	    (VT1732_CFG_SPDF_INT_TX | VT1732_CFG_SPDF_EXT_TX))
		chip->spdif_out = 1;
	if (chip->eeprom.data[EEPROM_SPDIF] & VT1732_CFG_SPDF_EXT_RX)
		chip->spdif_in = 1;

	return 0;
}


/*
 */

static int __devinit vt1732_chip_init(struct snd_vt1732 *chip)
{
	int err;

#if 0 /* 60-63 become invalid with envy24-II mode */
	pci_write_config_byte(chip->pci, 0x60, chip->eeprom.data[0]);
	pci_write_config_byte(chip->pci, 0x61, chip->eeprom.data[1]);
	pci_write_config_byte(chip->pci, 0x62, chip->eeprom.data[2]);
	pci_write_config_byte(chip->pci, 0x63, chip->eeprom.data[3]);
#endif

	/* power up everything */
	vt1732_writeb(chip, VT1732_POWER_DOWN, 0);
	vt1732_writeb(chip, VT1732_EXT_POWER_DOWN, 0);

	vt1732_writeb(chip, VT1732_CONFIG, chip->eeprom.data[0]);
	vt1732_writeb(chip, VT1732_INT_CONFIG, chip->eeprom.data[1]);
	vt1732_writeb(chip, VT1732_I2S_CONFIG, chip->eeprom.data[2]);
	vt1732_writeb(chip, VT1732_SPDIF_CONFIG, chip->eeprom.data[3]);

	vt1732_writeb(chip, VT1732_INT_CONFIG1, 0x90);

	snd_vt1732_set_rate(chip, chip->rate, 1);

	if (chip->config->init) {
		err = chip->config->init(chip);
		if (err < 0)
			return err;
	}

	return 0;
}

static int snd_vt1732_free(struct snd_vt1732 *chip)
{
	if (chip->iomap) {
		/* disable irqs */
		vt1732_writew(chip, VT1732_DMA_INT_MASK, -1);
		vt1732_writel(chip, VT1732_PLAY_INT_MASK, -1);
		vt1732_writel(chip, VT1732_REC_INT_MASK, -1);
		if (chip->config && chip->config->free)
			chip->config->free(chip);
		iounmap(chip->iomap);
	}
	if (chip->irq >= 0) {
		synchronize_irq(chip->irq);
		free_irq(chip->irq, chip);
	}
	pci_release_regions(chip->pci);
	pci_disable_device(chip->pci);
	kfree(chip);
	return 0;
}

static int snd_vt1732_dev_free(struct snd_device *device)
{
	struct snd_vt1732 *chip = device->device_data;
	return snd_vt1732_free(chip);
}

#define VT1732_RDMAx_REG(num, type) \
	(rdmax_base_ofs[num - 1] + VT1732_PDMAx_##type)


static unsigned short xdmax_base_ofs[2][VT1732_PCM_STREAMS] __devinitdata = {
	{
		VT1732_PDMA0_BASE_ADDR,		/* PDMA0 */
		VT1732_PDMA1_BASE_ADDR,		/* PDMA1 */
		VT1732_PDMA1_BASE_ADDR + 0x10,	/* PDMA2 */
		VT1732_PDMA1_BASE_ADDR + 0x20,	/* PDMA3 */
		VT1732_PDMA1_BASE_ADDR + 0x30,	/* PDMA4 */
		VT1732_PDMA1_BASE_ADDR + 0x40,	/* PDMA5 */
		VT1732_PDMA1_BASE_ADDR + 0x50,	/* PDMA6 */
		VT1732_PDMA1_BASE_ADDR + 0x40,	/* PDMA7 */
		VT1732_PDMA8_BASE_ADDR,		/* PDMA8 */
		VT1732_PDMA8_BASE_ADDR + 0x10,	/* PDMA9 */
		VT1732_PDMA8_BASE_ADDR + 0x20,	/* PDMA10 */
		VT1732_PDMA8_BASE_ADDR + 0x30,	/* PDMA11 */
		VT1732_PDMA8_BASE_ADDR + 0x40,	/* PDMA12 */
		VT1732_PDMA8_BASE_ADDR + 0x50,	/* PDMA13 */
		VT1732_PDMA8_BASE_ADDR + 0x60,	/* PDMA14 */
		VT1732_PDMA8_BASE_ADDR + 0x70,	/* PDMA15 */
		VT1732_PDMASP0_BASE_ADDR,	/* SPDIF */
		0
	},
	{
		VT1732_RDMA0_BASE_ADDR,		/* RMDA0 */
		VT1732_RDMA1_BASE_ADDR,		/* RDMA1 */
		VT1732_RDMA1_BASE_ADDR + 0x10,	/* RDMA2 */
		VT1732_RDMA1_BASE_ADDR + 0x20,	/* RDMA3 */
		VT1732_RDMA1_BASE_ADDR + 0x30,	/* RDMA4 */
		VT1732_RDMA1_BASE_ADDR + 0x40,	/* RDMA5 */
		VT1732_RDMA1_BASE_ADDR + 0x50,	/* RDMA6 */
		VT1732_RDMA1_BASE_ADDR + 0x40,	/* RDMA7 */
		VT1732_RDMA8_BASE_ADDR,		/* RDMA8 */
		VT1732_RDMA8_BASE_ADDR + 0x10,	/* RDMA9 */
		VT1732_RDMA8_BASE_ADDR + 0x20,	/* RDMA10 */
		VT1732_RDMA8_BASE_ADDR + 0x30,	/* RDMA11 */
		VT1732_RDMA8_BASE_ADDR + 0x40,	/* RDMA12 */
		VT1732_RDMA8_BASE_ADDR + 0x50,	/* RDMA13 */
		VT1732_RDMA8_BASE_ADDR + 0x60,	/* RDMA14 */
		VT1732_RDMA8_BASE_ADDR + 0x70,	/* RDMA15 */
		VT1732_RDMASP0_BASE_ADDR,	/* SPDIF */
		VT1732_RDMADM0_BASE_ADDR,	/* MIXER */
	}
};

static void __devinit vt1732_setup_streams(struct snd_vt1732 *chip)
{
	int s, i;

	for (s = 0; s < 2; s++) {
		/* [PR]DMA0-7: bit0-7 */
		for (i = 0; i < 8; i++)
			chip->str[s][i].stat_bit = (1 << i);
		/* [PR]DMA8-15: bit10-17 */
		for (; i < 16; i++)
			chip->str[s][i].stat_bit = (1 << (i + 2));
		/* [PR]DMASPDIF0: bit8 */
		chip->str[s][VT1732_PCM_SPDIF].stat_bit = (1 << 8);
		/* RDMADM0: bit18 */
		if (s)
			chip->str[s][VT1732_PCM_MIXER].stat_bit = (1 << 18);
		for (i = 0; i < VT1732_PCM_STREAMS; i++)
			chip->str[s][i].base_reg = xdmax_base_ofs[s][i];
	}
}

static int __devinit snd_vt1732_create(struct snd_card *card,
				       struct pci_dev *pci,
				       struct snd_vt1732 **rchip)
{
	struct snd_vt1732 *chip;
	int err;
	static struct snd_device_ops ops = {
		.dev_free =	snd_vt1732_dev_free,
	};

	*rchip = NULL;

	err = pci_enable_device(pci);
	if (err < 0)
		return err;

	chip = kzalloc(sizeof(*chip), GFP_KERNEL);
	if (!chip) {
		pci_disable_device(pci);
		return -ENOMEM;
	}

	spin_lock_init(&chip->reg_lock);
	mutex_init(&chip->open_mutex);
	chip->card = card;
	chip->pci = pci;
	chip->irq = -1;
	chip->rate = 48000;

	vt1732_setup_streams(chip);

	pci_set_master(pci);

	err = pci_request_regions(pci, "VT1732");
	if (err < 0) {
		kfree(chip);
		pci_disable_device(pci);
		return err;
	}

	/* 0: Envy24-compatible I/O base,
	 * 1: iomap
	 * 2: shadow map (read-only)
	 */
	chip->addr = pci_resource_start(pci, 1);
	chip->iomap = ioremap_nocache(chip->addr, pci_resource_len(pci, 1));

	if (request_irq(pci->irq, vt1732_interrupt, IRQF_SHARED,
			"VT1732", chip)) {
		snd_printk(KERN_ERR PFX "unable to grab IRQ %d\n", pci->irq);
		snd_vt1732_free(chip);
		return -EIO;
	}
	chip->irq = pci->irq;
	
	err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, chip, &ops);
	if (err < 0) {
		snd_vt1732_free(chip);
 		return err;
	}

	snd_card_set_dev(card, &pci->dev);
	*rchip = chip;
	return 0;
}

static void vt1732_proc_read(struct snd_info_entry *entry, 
			     struct snd_info_buffer *buffer)
{
	struct snd_vt1732 *chip = entry->private_data;
	int i;

	snd_iprintf(buffer, "Subvendor: %08x\n", chip->eeprom.subvendor);
	snd_iprintf(buffer, "EEPROM: %02x %02x %02x %02x\n\n",
		    chip->eeprom.data[0], chip->eeprom.data[1],
		    chip->eeprom.data[2], chip->eeprom.data[3]);
	for (i = 0; i < 0x400; i += 4) {
		int j;
		snd_iprintf(buffer, "%04x:", i);
		for (j = 0; j < 4; j++)
			snd_iprintf(buffer," %02x", vt1732_readb(chip, i + j));
		snd_iprintf(buffer, "\n");
	}
}

#ifdef CONFIG_SND_DEBUG
static void vt1732_proc_write(struct snd_info_entry *entry,
			      struct snd_info_buffer *buffer)
{
	struct snd_vt1732 *chip = entry->private_data;
	char line[64];
	unsigned int reg, val;

	while (!snd_info_get_line(buffer, line, sizeof(line))) {
		if (sscanf(line, "%x %x", &reg, &val) != 2)
			continue;
		if (reg < 0x400 && val < 0x100)
			vt1732_writeb(chip, reg, val);
	}
}
#endif

static void vt1732_proc_init(struct snd_vt1732 *chip)
{
	struct snd_info_entry *entry;

	if (!snd_card_proc_new(chip->card, "vt1732", &entry)) {
		snd_info_set_text_ops(entry, chip, vt1732_proc_read);
#ifdef CONFIG_SND_DEBUG
		entry->mode |= S_IWUSR;
		entry->c.text.write = vt1732_proc_write;
#endif
	}
}

static int __devinit snd_vt1732_probe(struct pci_dev *pci,
				      const struct pci_device_id *pci_id)
{
	static int dev;
	struct snd_card *card;
	struct snd_vt1732 *chip;
	int err;

	if (dev >= SNDRV_CARDS)
		return -ENODEV;
	if (!enable[dev]) {
		dev++;
		return -ENOENT;
	}

	err = snd_card_create(index[dev], id[dev], THIS_MODULE, 0, &card);
	if (err < 0)
		return err;

	strcpy(card->driver, "VT1732");
	strcpy(card->shortname, "VIA VT1732");
	
	err = snd_vt1732_create(card, pci, &chip);
	if (err < 0)
		goto error;

	vt1732_reset(chip);
	err = model_init(chip, model[dev]);
	if (err < 0)
		goto error;

	/* shortname might be changed in model_init */
	sprintf(card->longname, "%s at 0x%lx, irq %i",
		card->shortname, chip->addr, chip->irq);

	err = vt1732_chip_init(chip);
	if (err < 0)
		goto error;

	err = create_pcm_interleave(chip, 0);
	if (err < 0)
		goto error;
	err = create_pcm_stereo(chip, 1);
	if (err < 0)
		goto error;
	err = create_pcm_mono(chip, 2);
	if (err < 0)
		goto error;
	err = create_pcm_spdif(chip, 3);
	if (err < 0)
		goto error;
	err = create_pcm_mixer(chip, 4);
	if (err < 0)
		goto error;

	err = build_controls(chip);
	if (err < 0)
		goto error;

	err = create_midi(chip);
	if (err < 0)
		goto error;

	vt1732_proc_init(chip);

	err = snd_card_register(card);
	if (err < 0)
		goto error;

	pci_set_drvdata(pci, card);
	dev++;

	return 0;

 error:
	snd_card_free(card);
	return err;
}

static void __devexit snd_vt1732_remove(struct pci_dev *pci)
{
	snd_card_free(pci_get_drvdata(pci));
	pci_set_drvdata(pci, NULL);
}

#define PCI_DEVICE_ID_VT1732	0x1732

static struct pci_device_id snd_vt1732_ids[] __devinitdata = {
	/* { PCI_DEVICE(PCI_VENDOR_ID_VIA, PCI_DEVICE_ID_VT1732), }, */
	{ PCI_DEVICE(0x1106, 0x1732), },
	{ 0, }
};

MODULE_DEVICE_TABLE(pci, snd_vt1732_ids);

static struct pci_driver driver = {
	.name = "VT1732",
	.id_table = snd_vt1732_ids,
	.probe = snd_vt1732_probe,
	.remove = __devexit_p(snd_vt1732_remove),
};

static int __init alsa_card_vt1732_init(void)
{
	return pci_register_driver(&driver);
}

static void __exit alsa_card_vt1732_exit(void)
{
	pci_unregister_driver(&driver);
}

module_init(alsa_card_vt1732_init)
module_exit(alsa_card_vt1732_exit)
